program Appl4Loader;

(*
================================================================================================================
Author:
   This Application sample using the CPULoader for AVRco Pascal has been written Michael Bremicker Oct. 2009

================================================================================================================
Changes:

   Sept. 2012:
   
   - Main Application supports now a small Flash Writer Procedure in this Main Application to update the BootLoader Program. The binary code of the CPULoader update is read from the MMC Card using FAT16
   - The BootFlag Record contains now a Minimum Required Version - to make sure, that the current Main Application is fit for the planned update

   Oct. 2009:
   
   - Original Version
   
   Please note the {!} that are in the source - here you have the important settings you need for your own application!
================================================================================================================

*)


{$NOSHADOW}
{ $WG}                     {global Warnings off}
{!}
Device = mega2561, VCC=5;


Define_Fuses                                                                    // Set the Fuses here the way you need them in case you want to run the application without the CPULoader
  Override_Fuses;
  LockBits0      = [];
  FuseBits0      = [SUT1];
{!}
  FuseBits1      = [SPIEN, EESAVE, BOOTSZ1, BOOTSZ0];                           // Here do NOT set the BOOTRST Fuse, so the application will be started normally without the CPULoader

  FuseBits2      = [BODLEVEL0,BODLEVEL2];
  ProgMode       = SPI;

Import SysTick, FlashWrite, FAT16;                                              //

{!}
From System Import LongWord, FlashCheck_S;                                      // This is required for the compiler to calculate the CheckSum of the program

Define
{!}
  FlashChkSum    = $3E000-2;                                                    // Byte address of the location, where the compiler shall place the calculated CheckSum

  ProcClock      = 14745600;
  SysTick        = 10;             {msec}
  StackSize      = $0100, iData;
  FrameSize      = $0400, iData;

  FAT16           = MMC_SPI, IData;
  F16_FileHandles = 2;
  F16_DirLevels   = 2;
  F16_MMCspeed    = Fast;// standard, slow, fast


Uses uFAT16;

Implementation

{$IDATA}

{--------------------------------------------------------------}
{ Type Declarations }

Type
  tBootFlag           = Record                                                  // Definition of the BootFlag Record at the end of the application flash area
                          MinPrevVersion : WORD;                                // Minimum Version Number of the current Main Application, that is required for allowing updating
                          Version : WORD;
                          Flag    : WORD;
                        End;

{--------------------------------------------------------------}
{ Const Declarations }
Const
{!} // -->
   Boot_Start                : LONGWORD = $3F000;                               // Pointer to the first Byte of the CPULoader Program. The Boot-Area however begins here at $3E000;
   BootFlag[@Boot_Start - 8] : tBOOTFLAG =                                      // Now place the BootFlag Record 8 Bytes before the end. That leaves 2 bytes left for the FlashCkSum (see above)
                                           (MinPrevVersion : 0000;              // The minimum required Version of the current application (that shall be overwritten). 0000 Allows any update
                                            Version        : PROJECTBUILD+1;    // Version is the PROJECTBUILD number (plus 1 to have the same numbering as the AVRco compiler shows..)
                                            Flag           : $AA55);            // Flag is used to indicate to the bootloader, that a valid application resides in the application flash memory.
                                                                                // This Flag is flashed last! If the bootloader finds $AA55 it knows, that there is a valid application.
{!} // <--

{--------------------------------------------------------------}
{ Var Declarations }
{$IDATA}
Var
   gF        : File;
   gWordBuff : Array [0..127] Of Word;

{--------------------------------------------------------------}
{ functions / procedures }

Procedure JumpToCPULoader;                                                      // This jumps to the CPULoader. Do NOT forget to prepare your application for the RESET!
begin
   EXCL(MCUSR,1);                                                               // Clear the ResetFlag in case it is still there: Sometimes it does not get cleared after a Power On
   EXCL(MCUSR,0);                                                               // Clear the POWER ON ResetFlag
   mDelay(50);
   Asm;
      Cli                                                                       // Disable all interrupts
      Jmp Appl4Loader.Boot_Start                                                // and jump to the CPULoader
   EndAsm;
end;



Procedure CPULoaderUpdater;                                                     // This is the Updater Routine for the CPULoader.
                                                                                // PLEASE NOTE: This routine does NOT check the Version or the File Check sum of the LOADER.BIN file! You must handle that yourself.
                                                                                //              The LOADER.BIN file MUST be exactely 4 K of size! After compiling it and saving it to disk using the
                                                                                //              "Save Flash Buffer to binary File" in the File Menue of the AVRco programmer you must cut off all the "FF" bytes before the
                                                                                //              CPULoader Code begins at $3F000 manually, using a proper hex editor tool.
Var
   Res : Word;
   lError : Boolean;
begin
   F16_FileAssign(gF, '\','CPULOADER.BIN');                                     // Get the Code file of the new CPULoader Update from the MMC Card Root.
   If F16_FileReset (gF) Then                                                   //
      FlashInitPage(Addr(Boot_Start));                                          // Point to the start of the Bootloader, i.e. $3F000 (ByteAdr.) = $1F800 (Word Adr.): now FLASH_PAGE is 3 and FLASH_ADDR is $F000
      Repeat
         If F16_BlockRead(gF, @gWordBuff, SizeOf(gWordBuff), Res) Then          // Read the file in blocks of 256 bytes
            FlashErasePage;                                                     // Erase the current Flash-Page
            FlashCopyR2F(@gWordBuff);                                           // Write the Block from the RAM into the internal CPU-Buffer
            FlashProgPage;                                                      // And flash the internal buffer
            Inc(FLASH_ADDR,$100);                                               // And increment the pointer $100 = 256 bytes for the next page
         Else
            lError := True;
         EndIf;
      Until F16_EndOfFile(gF) Or lError;
   EndIf;
   F16_FileClose(gF);
End;



{$Validate FLASHERASEPAGE}                                                      // These procedures are for Flash-Writing the BootLoader from the Main Application
{$Validate FLASHPROGPAGE}                                                       // The Validate-Switch places them after the Boot-Block (see below) without optimizing
{$Validate FlashCopyF2R}
{$Validate FlashCopyR2F}

{$PHASE BootBlock $1F790}                                                       // Here we start the FlashLoader for the BootLoader, as far as possible to the End of the Application Flash Area
Procedure FlashBootBlock;                                                       // The Fuses BOOTSZ0/SZ1 make this area already a Boot-Loader Area, thus, the SPM Command for flashing can be executed!
Begin                                                                           // This procedure is a DUMMY-Call for enforcing AVRco to put the above validated routines to this address
//Asm: NOP;                                                                     // Have a look into the .LST file, that is generated by the AVRco compiler.
End;
{$DEPHASE BootBlock}                                                            // End of FlashLoader BootBlock - Make sure, it does NOT overwrite the BootFlag Record and the Checksum!!


{--------------------------------------------------------------}
{ Main Program }
{$IDATA}

Begin
  EnableInts;
  mDelay (5000);

  CPULoaderUpdater;                                                             // This procedure will read in a binary file for updating the CPULoader itself. The Main Application stays untouched.

  mDelay (5000);

  JumpToCPULoader;                                                              // This procedure will leave the application and invoke the CPULoader - which will update this Main Application
  

  loop


  endloop;
end Appl4Loader.

